"""Jeff's Unified Registration Tool

Facilitate coregistration and spatial normalization of fMRI datasets.
"""
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

# Set the version string
# This is automatically updated by bumpversion (see .bumpversion.cfg)
__version__ = '1.0.1'

# Import classes from submodules into the jurt namespace
from jurt.core import Prefix
from jurt.core import Dataset
from jurt.core import Pipeline
from jurt.core import FsPipeline
from jurt.core import FslPipeline
from jurt.t1 import PrepT1
from jurt.t1 import SST1
from jurt.t1 import SegT1
from jurt.reg import PrepFunc
from jurt.reg import FuncToT1
from jurt.reg import T1ToStd
from jurt.reg import FuncToStd

###############################################################################

if __name__ == '__main__':
    raise RuntimeError('jurt/__init__.py cannot be directly executed')

###############################################################################




"""Unit tests for jurt"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import unittest
import sys
import os
import tempfile
import logging
import jurt

###############################################################################

class TestPrefix(unittest.TestCase):
    """Test the jurt.Prefix class"""

    def test_prefix_noarg(self):
        with self.assertRaisesRegex(TypeError,
            "__new__\(\) missing 1 required positional argument: 'prefix'"):
            obj = jurt.Prefix()

    def test_prefix_none(self):
        with self.assertRaisesRegex(TypeError, 'prefix must be a string'):
            obj = jurt.Prefix(None)

    def test_prefix_empty(self):
        with self.assertRaisesRegex(ValueError, 'prefix cannot be empty'):
            obj = jurt.Prefix('')

    def test_prefix_bad_dir(self):
        td = tempfile.TemporaryDirectory()
        td_name = td.name
        td.cleanup()
        self.assertFalse(os.path.isdir(td_name))
        with self.assertRaisesRegex(IOError,
            f'Could not find directory: {td_name}'):
            obj = jurt.Prefix(os.path.join(td_name, 'TEST'))

    def test_prefix_good_without_path(self):
        test_prefix = 'TEST'
        obj = jurt.Prefix(test_prefix)
        self.assertEqual(obj, os.path.join(os.getcwd(), test_prefix))

    def test_prefix_good_with_path(self):
        with tempfile.TemporaryDirectory() as td:
            test_prefix = os.path.join(td, 'TEST')
            obj = jurt.Prefix(test_prefix)
            self.assertEqual(obj, test_prefix)

    def test_prefix_reassign(self):
        test_prefix = 'TEST01'
        obj = jurt.Prefix(test_prefix)
        self.assertEqual(obj, os.path.join(os.getcwd(), test_prefix))
        with self.assertRaisesRegex(TypeError, 'prefix must be a string'):
            obj = jurt.Prefix(None)
        self.assertEqual(obj, os.path.join(os.getcwd(), test_prefix))
        with self.assertRaisesRegex(ValueError, 'prefix cannot be empty'):
            obj = jurt.Prefix('')
        self.assertEqual(obj, os.path.join(os.getcwd(), test_prefix))
        test_prefix = 'TEST02'
        obj = jurt.Prefix(test_prefix)
        self.assertEqual(obj, os.path.join(os.getcwd(), test_prefix))

###############################################################################

class TestDataset(unittest.TestCase):
    """Test the jurt.Dataset class"""

    def test_dataset_noarg(self):
        with self.assertRaisesRegex(TypeError,
            "__new__\(\) missing 1 required positional argument: 'dataset'"):
            obj = jurt.Dataset()

    def test_dataset_none(self):
        with self.assertRaisesRegex(TypeError, 'dataset name must be a string'):
            obj = jurt.Dataset(None)

    def test_dataset_empty(self):
        with self.assertRaisesRegex(ValueError, 'dataset name cannot be empty'):
            obj = jurt.Dataset('')

    def test_dataset_bad(self):
        with tempfile.TemporaryDirectory() as td:
            tf = os.path.join(td, 'test.nii')
            with self.assertRaisesRegex(IOError, f'Could not find dataset: {tf}'):
                obj = jurt.Dataset(tf)

    def test_dataset_good(self):
        with tempfile.NamedTemporaryFile() as tf:
            obj = jurt.Dataset(tf.name)
            self.assertEqual(obj, tf.name)

    def test_dataset_reassign(self):
        with tempfile.NamedTemporaryFile() as tf1:
            obj = jurt.Dataset(tf1.name)
            self.assertEqual(obj, tf1.name)
            with self.assertRaisesRegex(TypeError, 'dataset name must be a string'):
                obj = jurt.Dataset(None)
            self.assertEqual(obj, tf1.name)
            with self.assertRaisesRegex(ValueError, 'dataset name cannot be empty'):
                obj = jurt.Dataset('')
            self.assertEqual(obj, tf1.name)
            with tempfile.NamedTemporaryFile() as tf2:
                obj = jurt.Dataset(tf2.name)
                self.assertEqual(obj, tf2.name)

###############################################################################

class TestPipeline(unittest.TestCase):
    """Test the jurt.Pipeline class"""

    def setUp(self):
        # Use a NullHandler for logging for this round of tests to prevent
        # extra console output when expected exceptions are raised
        logging.getLogger('jurt').addHandler(logging.NullHandler())

        # Set up the jurt.Pipeline() object
        self.obj = jurt.Pipeline()
        self.assertIsNotNone(self.obj)
        self.assertIsNone(self.obj.prefix)
        self.assertFalse(self.obj.overwrite)
        self.assertIsNone(self.obj.scratch)
        self.assertEqual(self.obj.threads, jurt.Pipeline._default_threads)
        self.assertEqual(self.obj.umask, jurt.Pipeline._default_umask)

    def test_overwrite_none(self):
        with self.assertRaisesRegex(TypeError, 'overwrite must be boolean'):
            self.obj.overwrite = None

    def test_overwrite_false(self):
        self.obj.overwrite = False
        self.assertFalse(self.obj.overwrite)

    def test_overwrite_true(self):
        self.obj.overwrite = True
        self.assertTrue(self.obj.overwrite)

    def test_scratch_none(self):
            self.obj.scratch = None
            self.assertIsNone(self.obj.scratch)

    def test_scratch_bad_type(self):
        with self.assertRaisesRegex(TypeError,'scratch must be a string or None'):
            self.obj.scratch = 0

    def test_scratch_empty(self):
        with self.assertRaisesRegex(ValueError,'scratch cannot be empty'):
            self.obj.scratch = ''

    def test_scratch_bad_dir(self):
        with tempfile.TemporaryDirectory() as td:
            d = os.path.join(td, 'TEST')
            with self.assertRaisesRegex(IOError, f'Could not find directory: {d}'):
                self.obj.scratch = d

    def test_scratch_good_dir(self):
        with tempfile.TemporaryDirectory() as td:
            self.obj.scratch = td
            self.assertEqual(self.obj.scratch, td)

    def test_threads_none(self):
        with self.assertRaisesRegex(TypeError, 'threads must be an integer'):
            self.obj.threads = None

    def test_threads_string(self):
        with self.assertRaisesRegex(TypeError, 'threads must be an integer'):
            self.obj.threads = '8'

    def test_threads_negative(self):
        with self.assertRaisesRegex(ValueError, 'threads must be greater than zero'):
            self.obj.threads = -8

    def test_threads_zero(self):
        with self.assertRaisesRegex(ValueError, 'threads must be greater than zero'):
            self.obj.threads = 0

    def test_threads_positive(self):
        self.obj.threads = 8
        self.assertEqual(self.obj.threads, 8)

    def test_umask_none(self):
        with self.assertRaisesRegex(TypeError, 'umask must be an integer'):
            self.obj.umask = None

    def test_umask_string(self):
        with self.assertRaisesRegex(TypeError, 'umask must be an integer'):
            self.obj.umask = '007'

    def test_umask_invalid(self):
        with self.assertRaisesRegex(ValueError,
            'umask must be in the range \[%#05o,%#05o\]' % (0, 0o777)):
            self.obj.umask = 0o7771

    def test_umask_valid(self):
        umask = 0o007
        self.obj.umask = umask
        self.assertEqual(self.obj.umask, umask)

    def test_run_no_prefix(self):
        with self.assertRaisesRegex(RuntimeError,
            f'prefix must be set prior to {type(self.obj).__name__}.run\(\)'):
            self.obj.run()

    def test_run(self):
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(NotImplementedError,
            f'{type(self.obj).__name__}.pipeline\(\) must be overridden'):
            self.obj.run()

    def test_cmd_redirect(self):
        teststr = 'Hello, world!'
        with tempfile.TemporaryDirectory() as td:
            fn = os.path.join(td, 'test.log')
            with open(fn, 'w', buffering=1) as f:
                self.obj._cmd('echo', teststr, redirect=f)
            with open(fn, 'r') as f:
                self.assertEqual(f.readline().strip(), 'echo ' + teststr)
                self.assertEqual(f.readline().strip(), teststr)

###############################################################################

class TestFsPipeline(unittest.TestCase):
    """Test the jurt.FsPipeline class"""

    def setUp(self):
        self.obj = jurt.FsPipeline()
        self.assertIsNotNone(self.obj)

    def test_run(self):
        logging.getLogger('jurt').addHandler(logging.NullHandler())
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(NotImplementedError,
            f'{type(self.obj).__name__}.pipeline\(\) must be overridden'):
            self.obj.run()

###############################################################################

class TestFslPipeline(unittest.TestCase):
    """Test the jurt.FslPipeline class"""

    def setUp(self):
        self.obj = jurt.FslPipeline()
        self.assertIsNotNone(self.obj)

    def test_run(self):
        logging.getLogger('jurt').addHandler(logging.NullHandler())
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(NotImplementedError,
            f'{type(self.obj).__name__}.pipeline\(\) must be overridden'):
            self.obj.run()

###############################################################################

class TestPrepT1(unittest.TestCase):
    """Test the jurt.PrepT1 class"""

    def setUp(self):
        self.obj = jurt.PrepT1()
        self.assertIsNotNone(self.obj)
        self.assertIsNone(self.obj.raw)

    def test_run_no_raw(self):
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(RuntimeError,
            f'raw must be set prior to {type(self.obj).__name__}.run\(\)'):
            self.obj.run()

###############################################################################

class TestSST1(unittest.TestCase):
    """Test the jurt.SST1 class"""

    def setUp(self):
        self.obj = jurt.SST1()
        self.assertIsNotNone(self.obj)
        self.assertEqual(self.obj.preflood, jurt.SST1._preflood_default)
        self.assertEqual(self.obj.gcut,  jurt.SST1._gcut_default)

    def test_preflood_none(self):
        with self.assertRaisesRegex(TypeError,
            'preflood must be an integer'):
            self.obj.preflood = None

    def test_preflood_bool(self):
        with self.assertRaisesRegex(TypeError,
            'preflood must be an integer'):
            self.obj.preflood = True

    def test_preflood_string(self):
        with self.assertRaisesRegex(TypeError,
            'preflood must be an integer'):
            self.obj.preflood = '25'

    def test_preflood_too_low(self):
        with self.assertRaisesRegex(ValueError,
            'preflood must be in the range \[0,100\]'):
            self.obj.preflood = -1

    def test_preflood_too_high(self):
        with self.assertRaisesRegex(ValueError,
            'preflood must be in the range \[0,100\]'):
            self.obj.preflood = 101

    def test_preflood_good(self):
        self.obj.preflood = 35
        self.assertEqual(self.obj.preflood, 35)

    def test_gcut_none(self):
        with self.assertRaisesRegex(TypeError,
            'gcut must be numeric'):
            self.obj.gcut = None

    def test_gcut_bool(self):
        with self.assertRaisesRegex(TypeError,
            'gcut must be numeric'):
            self.obj.gcut = True

    def test_gcut_string(self):
        with self.assertRaisesRegex(TypeError,
            'gcut must be numeric'):
            self.obj.gcut = '.4'

    def test_gcut_zero(self):
        self.obj.gcut = 0
        self.assertEqual(self.obj.gcut, 0.0)

    def test_gcut_too_low(self):
        with self.assertRaisesRegex(ValueError,
            'gcut must be in the range \[0\.0,1\.0\)'):
            self.obj.gcut = -0.1

    def test_thresh_too_high(self):
        with self.assertRaisesRegex(ValueError,
            'gcut must be in the range \[0\.0,1\.0\)'):
            self.obj.gcut = 1.0

    def test_thresh_good(self):
        self.obj.gcut = 0.36
        self.assertEqual(self.obj.gcut, 0.36)

###############################################################################

class TestSegT1(unittest.TestCase):
    """Test the jurt.SegT1 pipeline"""

    def test_seg_t1(self):
        obj = jurt.SegT1()
        self.assertIsNotNone(obj)

###############################################################################

class TestPrepFunc(unittest.TestCase):
    """Test the jurt.PrepFunc pipeline"""

    def setUp(self):
        self.obj = jurt.PrepFunc()
        self.assertIsNotNone(self.obj)
        self.assertIsNone(self.obj.func)
        self.assertIsNone(self.obj.basevol)

    def test_basevol_none(self):
        with self.assertRaisesRegex(TypeError, 'basevol must be an integer'):
            self.obj.basevol = None

    def test_basevol_string(self):
        with self.assertRaisesRegex(TypeError, 'basevol must be an integer'):
            self.obj.basevol = '0'

    def test_basevol_negative(self):
        with self.assertRaisesRegex(ValueError, 'basevol cannot be negative'):
            self.obj.basevol = -1

    def test_basevol_good(self):
        self.obj.basevol = 0
        self.assertEqual(self.obj.basevol, 0)

    def test_run_no_func(self):
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(RuntimeError,
            f'func must be set prior to {type(self.obj).__name__}.run\(\)'):
            self.obj.run()

    def test_run_no_basevol(self):
        self.obj.prefix = 'TEST'
        with tempfile.NamedTemporaryFile() as tf:
            self.obj.func = tf.name
            with self.assertRaisesRegex(RuntimeError,
                f'basevol must be set prior to {type(self.obj).__name__}.run\(\)'):
                self.obj.run()

###############################################################################

class TestFuncToT1(unittest.TestCase):
    """Test the jurt.FuncToT1 pipeline"""

    def setUp(self):
        self.obj = jurt.FuncToT1()
        self.assertIsNotNone(self.obj)
        self.assertIsNone(self.obj.method)

    def test_dof_none(self):
        with self.assertRaisesRegex(TypeError, 'method must be a string'):
            self.obj.method = None

    def test_method_bad(self):
        with self.assertRaisesRegex(ValueError,
            'method must be one of: {0:s}'.format(
            ', '.join(jurt.FuncToT1._valid_methods))):
            self.obj.method = 'TEST'

    def test_method_bbr(self):
        self.obj.method = 'bbr'
        self.assertEqual(self.obj.method, 'bbr')

    def test_method_6dof(self):
        self.obj.method = '6dof'
        self.assertEqual(self.obj.method, '6dof')

    def test_run_no_method(self):
        logging.getLogger('jurt').addHandler(logging.NullHandler())
        self.obj.prefix = 'TEST'
        with self.assertRaisesRegex(RuntimeError,
            f'method must be set prior to {type(self.obj).__name__}.run()'):
            self.obj.run()

###############################################################################

class TestT1ToStd(unittest.TestCase):
    """Test the jurt.T1ToStd pipeline"""

    def test_t1_to_std(self):
        obj = jurt.T1ToStd()
        self.assertIsNotNone(obj)

###############################################################################

class TestFuncToStd(unittest.TestCase):
    """Test the jurt.FuncToStd pipeline"""

    def test_func_to_std(self):
        obj = jurt.FuncToStd()
        self.assertIsNotNone(obj)

###############################################################################

if __name__ == '__main__':
    unittest.main()

###############################################################################




"""Entry point for jurt"""
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import sys
import argparse
import logging
import logging.handlers
import jurt

def main(argv=None):
    """Main jurt routine"""

    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser(
        prog='jurt',
        add_help=False,
        allow_abbrev=False,
        usage='jurt [options] <method> ...',
        description='Coregister and spatially normalize fMRI datasets.',
        epilog="""
        For help on a particular method, use 'jurt <method> -help'.
        """)
    g1 = parser.add_argument_group('Options')
    g1.add_argument('-loglevel',
            choices=('DEBUG', 'INFO', 'WARN', 'ERROR') , default='WARN',
            help='Log level')
    g1.add_argument('-logfile', metavar='file',
            help='Log to file (instead of console)')
    g1.add_argument('-version',
        action='version',
        version=f'jurt-{jurt.__version__}',
        help='Show version number and exit')
    g1.add_argument('-help',
        action='help',
        help='Show this help message and exit')

    subparsers = parser.add_subparsers(title='Methods')

    prep_t1 = subparsers.add_parser('prep_t1',
        add_help=False,
        allow_abbrev=False,
        usage='jurt prep_t1 ...',
        parents=[jurt.PrepT1.parser()],
        help=jurt.PrepT1.__doc__.splitlines()[0],
        description=jurt.PrepT1.__doc__.splitlines()[0],
        epilog=''.join(jurt.PrepT1.__doc__.splitlines()[2:]))
    prep_t1.set_defaults(main=jurt.PrepT1.main)

    ss_t1 = subparsers.add_parser('ss_t1',
        add_help=False,
        allow_abbrev=False,
        usage='jurt ss_t1 ...',
        parents=[jurt.SST1.parser()],
        help=jurt.SST1.__doc__.splitlines()[0],
        description=jurt.SST1.__doc__.splitlines()[0],
        epilog=''.join(jurt.SST1.__doc__.splitlines()[2:]))
    ss_t1.set_defaults(main=jurt.SST1.main)

    seg_t1 = subparsers.add_parser('seg_t1',
        add_help=False,
        allow_abbrev=False,
        usage='jurt seg_t1 ...',
        parents=[jurt.SegT1.parser()],
        help=jurt.SegT1.__doc__.splitlines()[0],
        description=jurt.SegT1.__doc__.splitlines()[0],
        epilog=''.join(jurt.SegT1.__doc__.splitlines()[2:]))
    seg_t1.set_defaults(main=jurt.SegT1.main)

    prep_func = subparsers.add_parser('prep_func',
        add_help=False,
        allow_abbrev=False,
        usage='jurt prep_func ...',
        parents=[jurt.PrepFunc.parser()],
        help=jurt.PrepFunc.__doc__.splitlines()[0],
        description=jurt.PrepFunc.__doc__.splitlines()[0],
        epilog=''.join(jurt.PrepFunc.__doc__.splitlines()[2:]))
    prep_func.set_defaults(main=jurt.PrepFunc.main)

    func_to_t1 = subparsers.add_parser('func_to_t1',
        add_help=False,
        allow_abbrev=False,
        usage='jurt func_to_t1 ...',
        parents=[jurt.FuncToT1.parser()],
        help=jurt.FuncToT1.__doc__.splitlines()[0],
        description=jurt.FuncToT1.__doc__.splitlines()[0],
        epilog=''.join(jurt.FuncToT1.__doc__.splitlines()[2:]))
    func_to_t1.set_defaults(main=jurt.FuncToT1.main)

    t1_to_std = subparsers.add_parser('t1_to_std',
        add_help=False,
        allow_abbrev=False,
        usage='jurt t1_to_std ...',
        parents=[jurt.T1ToStd.parser()],
        help=jurt.T1ToStd.__doc__.splitlines()[0],
        description=jurt.T1ToStd.__doc__.splitlines()[0],
        epilog=''.join(jurt.T1ToStd.__doc__.splitlines()[2:]))
    t1_to_std.set_defaults(main=jurt.T1ToStd.main)

    func_to_std = subparsers.add_parser('func_to_std',
        add_help=False,
        allow_abbrev=False,
        usage='jurt func_to_std ...',
        parents=[jurt.FuncToStd.parser()],
        help=jurt.FuncToStd.__doc__.splitlines()[0],
        description=jurt.FuncToStd.__doc__.splitlines()[0],
        epilog=''.join(jurt.FuncToStd.__doc__.splitlines()[2:]))
    func_to_std.set_defaults(main=jurt.FuncToStd.main)

    if len(argv) == 0:
        parser.print_help()
        return

    ns = parser.parse_args(argv)

    # Set up logging
    loglevel = getattr(logging, ns.loglevel)
    delattr(ns, 'loglevel')

    handler = None
    if 'logfile' in ns and ns.logfile is not None:
        handler = logging.handlers.WatchedFileHandler(ns.logfile)
    else:
        handler = logging.StreamHandler()

    delattr(ns, 'logfile')

    handler.setFormatter(logging.Formatter(logging.BASIC_FORMAT))
    logger = logging.getLogger('jurt')
    logger.setLevel(loglevel)
    logger.addHandler(handler)

    # Call whatever routine was selected
    ns.main(ns)

if __name__ == '__main__':
    main()




"""Unit test module for jurt"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import os
import unittest

def get_suite():
    """Return a unittest.TestSuite object for jurt"""
    loader = unittest.TestLoader()
    suite = loader.discover(os.path.abspath(os.path.dirname(__file__)))
    return suite




"""jurt's T1 module

Facilitate preprocessing and brain-extraction of T1-weighted datasets.
"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import os
import shutil
import argparse
import tempfile
import logging

import jurt
import jurt.core

###############################################################################

class PrepT1(jurt.core.FsPipeline):
    """Preprocess T1-weighted datasets using FreeSurfer

    Preprocessing is run via FreeSurfer's recon-all and consists of conforming
    the data to FreeSurfer orientation, non-uniformity correction, and
    intensity normalization.
    """

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use PrepT1 objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Required parameters')
        g1.add_argument('-raw', required=True, metavar='DSET',
            help=PrepT1.raw.__doc__)
        g2 = p.add_argument_group('Options')
        g2.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the PrepT1 pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

        # Set defaults
        self._raw    = None     # Raw T1-weighted dataset

    def __str__(self):
        return super().__str__() + '\n  %-20s : %s' % ('Raw dataset', self._raw)

    @property
    def prefix(self):
        """Output dataset prefix (including path)"""
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = jurt.core.Prefix(prefix)

    @property
    def raw(self):
        """Raw T1-weighted dataset"""
        return self._raw

    @raw.setter
    def raw(self, raw):
        self._raw = jurt.core.Dataset(raw)

    def pipeline(self):

        if self.raw is None:
            raise RuntimeError(
                f'raw must be set prior to {type(self).__name__}.run()')

        # Create a temporary directory for SUBJECTS_DIR
        wd = os.getcwd()
        try:
            with tempfile.TemporaryDirectory(dir=self._scratch) as sd:
                os.chdir(sd)
                self._log.debug(f'Temporary SUBJECTS_DIR: {sd}')

                # Get the subject ID from the prefix
                sid = os.path.basename(self._prefix)
                self._log.info(f'Subject ID: {sid}')

                # Map of FreeSurfer output to PrepT1 output
                orig = f'{sd}/{sid}/mri/orig.mgz'
                nu   = f'{sd}/{sid}/mri/nu.mgz'
                t1   = f'{sd}/{sid}/mri/T1.mgz'
                xfm  = f'{sd}/{sid}/mri/transforms/talairach.xfm'
                lta  = f'{sd}/{sid}/mri/transforms/talairach-with-skull.lta'
                log  = f'{sd}/{sid}/scripts/recon-all.log'

                outfiles = {
                    orig: f'{self._prefix}-t1.mgz',
                    nu:   f'{self._prefix}-t1-nu.mgz',
                    t1:   f'{self._prefix}-t1-inorm.mgz',
                    xfm:  f'{self._prefix}-t1-talairach.xfm',
                    lta:  f'{self._prefix}-t1-talairach-with-skull.lta',
                    log:  f'{self._prefix}-PrepT1.log'}


                # Check if output already exists (delete if overwriting)
                for k, v in outfiles.items():
                    self._log.debug(f'{k} --> {v}')
                    if os.path.isfile(v):
                        if self._overwrite:
                            os.remove(v)
                        else:
                            raise IOError(f'{v} already exists')

                # Run recon-all through intensity normalization
                self._cmd('recon-all',
                    '-motioncor',
                    '-nuintensitycor',
                    '-talairach',
                    '-normalization',
                    '-umask', '{0:03o}'.format(self._umask),
                    '-openmp', str(self._threads),
                    '-sd', sd,
                    '-s', sid,
                    '-i', self._raw)

                # Run mri_em_register, appending to the log file
                env = os.environ
                env['OMP_NUM_THREADS'] = str(self._threads)
                with open(log, 'a', buffering=1) as lf:
                    self._cmd('mri_em_register',
                        '-skull',
                        nu,
                        self._gca,
                        lta,
                        env=env,
                        redirect=lf)

                try:
                    # Copy output to destination
                    for k, v in outfiles.items():
                        shutil.copy2(k, v)

                    # Re-link XFM to MGZ
                    for f in (orig, nu, t1):
                        self._cmd('mri_add_xform_to_header',
                            '-c', outfiles[xfm], outfiles[f])

                except:
                    for k, v in outfiles.items():
                        if os.path.isfile(v):
                            os.remove(v)
                    raise
        except:
            raise
        finally:
            os.chdir(wd)

###############################################################################

class SST1(jurt.core.FsPipeline):
    """Skull strip T1-weighted datasets

    Removes non-brain tissue from T1-weighted datasets using FreeSurfer's
    Hybrid Watershed Algorithm (HWA) and by optionally applying graph cuts.
    """

    _preflood_default = 25
    _gcut_default     = 0.0
    _gcut_const       = 0.4

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts taht use SST1 objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Options')
        g1.add_argument('-preflood', metavar='h',
            type=int, default=SST1._preflood_default,
            help=SST1.preflood.__doc__)
        g1.add_argument('-gcut', metavar='gthresh', type=float, nargs='?',
            default=SST1._gcut_default,
            const=SST1._gcut_const,
            help=SST1.gcut.__doc__)
        g1.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the SST1 pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

        # Preflood height (-h) for mri_watershed
        self._preflood = SST1._preflood_default
        self._gcut     = SST1._gcut_default

    def __str__(self):
        return super().__str__() + (
            '\n  %-20s : %s\n  %-20s : %s' % (
                'preflood', self._preflood,
                'gcut'    , self._gcut))

    @property
    def preflood(self):
        """Preflooding height for the hybrid watershed algorithm"""
        return self._preflood

    @preflood.setter
    def preflood(self, preflood):
        # Default is 25
        # Range: 0-100
        # Lower values shrink the skull surface
        # Higher values expand the skull surfice

        if type(preflood) is not int:
            raise TypeError('preflood must be an integer')
        if preflood < 0 or preflood > 100:
            raise ValueError('preflood must be in the range [0,100]')
        self._preflood = preflood

    @property
    def gcut(self):
        """gcut threshold (percent of white matter intensity; 0 = no gcut)"""
        return self._gcut

    @gcut.setter
    def gcut(self, gcut):
        # Default is 0.4
        # Range: (0-1) non-inclusive, 0.0 means don't use gcut
        if type(gcut) is int:
            gcut = float(gcut)
        if type(gcut) is not float:
            raise TypeError('gcut must be numeric')
        if gcut < 0.0 or gcut >= 1.0:
            raise ValueError('gcut must be in the range [0.0,1.0)')
        self._gcut = gcut


    def pipeline(self):

        # Check input files
        inorm = jurt.core.Dataset(self._prefix + '-t1-inorm.mgz')
        lta   = jurt.core.Dataset(self._prefix + '-t1-talairach-with-skull.lta')

        # Check if output already exists (delete if overwriting)
        inorm_brain = self._prefix + '-t1-inorm-brain.mgz'
        log         = self._prefix + '-SST1.log'
        outfiles    = (inorm_brain, log)

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            with open(log, 'w', buffering=1) as lf:
                with tempfile.TemporaryDirectory(dir=self._scratch) as td:

                    inorm_hwa = inorm_brain
                    if self._gcut > 0.0:
                        inorm_hwa = os.path.join(td, 'hwa.nii')

                    # Run mri_watershed
                    self._cmd('mri_watershed',
                        '-T1',
                        '-h', str(self._preflood),
                        '-brain_atlas', self._gca, lta,
                        inorm,
                        inorm_hwa,
                        env=env,
                        redirect=lf)

                    if self._gcut > 0.0:
                        # Run gcut
                        if not os.path.isfile(inorm_hwa):
                            raise IOError(f'Could not find dataset: {inorm_hwa}')

                        self._cmd('mri_gcut',
                            '-110',
                            '-T', str(self._gcut),
                            '-mult', inorm_hwa,
                            inorm,
                            inorm_brain,
                            env=env,
                            redirect=lf)

        except:
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise

###############################################################################

class SegT1(jurt.core.FsPipeline, jurt.core.FslPipeline):
    """Segment T1-weighted dataset using FSL's FAST

    This pipeline uses FSL's FAST to segment the brain-extracted T1-weighted
    dataset, keeping the white-matter map for use in coregistration.

    FreeSurfer's mri_mask is used to apply the -t1-inorm-brain.mgz mask to the
    -t1-nu.mgz dataset and this masked, non-uniformity corrected dataset is
    used for segmentation.
    """

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use SegT1 objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
                parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Options')
        g1.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the SegT1 pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        jurt.core.FsPipeline.__init__(self)
        jurt.core.FslPipeline.__init__(self)

    def __str__(self):
        return super().__str__()

    def pipeline(self):

        # Check the input datasets
        nu_mgz    = jurt.core.Dataset(self._prefix + '-t1-nu.mgz')
        brainmask = jurt.core.Dataset(self._prefix + '-t1-inorm-brain.mgz')

        # Check if the output already exists (delete if overwriting)
        nu_nii       = self._prefix + '-t1-nu.nii'
        nu_brain_nii = self._prefix + '-t1-nu-brain.nii'
        wmseg_nii    = self._prefix + '-t1-nu-brain-wmseg.nii'
        wmedge_nii   = self._prefix + '-t1-nu-brain-wmedge.nii'
        log          = self._prefix + '-SegT1.log'
        outfiles     = (nu_nii, nu_brain_nii, wmseg_nii, wmedge_nii, log)

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        wd = os.getcwd()
        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            env['FSLOUTPUTTYPE'] = 'NIFTI'
            with open(log, 'w', buffering=1) as lf:
                with tempfile.TemporaryDirectory(dir=self._scratch) as td:
                    os.chdir(td)
                    self._log.debug(f'Temporary directory: {td}')

                    # Apply the brainmask to the nu dataset
                    nu_brain_mgz = os.path.join(td, 'nu-brain.mgz')
                    self._cmd('mri_mask',
                        nu_mgz,
                        brainmask,
                        nu_brain_mgz,
                        env=env,
                        redirect=lf)

                    # Convert nu_mgz and nu_brainmask_mgz to NIfTI
                    # Reorient to standard FSL format

                    for i, o, t in zip(
                        (nu_mgz,        nu_brain_mgz),
                        (nu_nii,        nu_brain_nii),
                        ('tmp-nu.nii', 'tmp-nu-brain.nii')):

                        tf = os.path.join(td, t)
                        self._cmd('mri_convert',
                            '-odt', 'float',
                            i,
                            tf,
                            env=env,
                            redirect=lf)

                        self._cmd('fslreorient2std', tf, o, env=env, redirect=lf)

                    # Do the segmentation
                    fast_dset = 'brain-fast'
                    self._cmd('fast',
                        '-v',
                        '-o', fast_dset,
                        nu_brain_nii,
                        env=env,
                        redirect=lf)

                    # Get the white matter map
                    wmseg = 'wmseg'
                    self._cmd('fslmaths',
                        fast_dset + '_pve_2',
                        '-thr', str(0.5),
                        '-bin',
                        wmseg,
                        env=env,
                        redirect=lf)

                    # Create the white matter edge map
                    wmedge = 'wmedge'
                    self._cmd('fslmaths',
                        wmseg,
                        '-edge',
                        '-bin',
                        '-mas',
                        wmseg,
                        wmedge,
                        env=env,
                        redirect=lf)

                    # Copy output to destination
                    shutil.copy2(os.path.join(td, wmseg + '.nii'),  wmseg_nii)
                    shutil.copy2(os.path.join(td, wmedge + '.nii'), wmedge_nii)

        except:
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise
        finally:
            os.chdir(wd)

###############################################################################

if __name__ == '__main__':
    raise RuntimeError('jurt/t1.py cannot be directly executed')

###############################################################################




"""jurt's Registration module

Unified motion correction, coregistration, and spatial normalization.
"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import os
import shutil
import argparse
import tempfile
import logging

import jurt
import jurt.core

###############################################################################

def nvols(dset):
    """Return the number of volumes in a dataset"""
    import subprocess

    if not os.path.isfile(dset):
        raise IOError(f'Could not find dataset: {dset}')
    cmd = subprocess.Popen(['fslnvols', dset], stdout=subprocess.PIPE)
    o, e = cmd.communicate()
    if cmd.returncode != 0:
        raise RuntimeError('fslnvols failed')
    return int(o.decode().splitlines()[0])

###############################################################################

class PrepFunc(jurt.core.FslPipeline):
    """Preprocess functional dataset

    Convert the functional dataset to FSL's preferred orientation and data
    type, extract the base volume, and run motion-correction using FSL's
    MCFLIRT.
    """

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use PrepFunc objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Required parameters')
        g1.add_argument('-func', required=True, metavar='DSET',
            help=PrepFunc.func.__doc__)
        g1.add_argument('-basevol', required=True, metavar='v', type=int,
            help=PrepFunc.basevol.__doc__)
        g2 = p.add_argument_group('Options')
        g2.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the PrepFunc pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

        # Set defaults
        self._func    = None
        self._basevol = None
        self._nvols   = 0       # Number of volumes in _func

    def __str__(self):
        return super().__str__() + (
            '\n  %-20s : %s\n  %-20s : %s\n  %-20s : %d' % (
            'Functional dataset', self._func,
            'Base volume', self._basevol,
            'fMRI volumes', self._nvols))

    @property
    def func(self):
        """Functional dataset"""
        return self._func

    @func.setter
    def func(self, func):
        self._func = jurt.core.Dataset(func)
        self._nvols = nvols(func)

    @property
    def basevol(self):
        """Registration base volume (functional dataset)"""
        return self._basevol

    @basevol.setter
    def basevol(self, basevol):
        if type(basevol) is not int:
            raise TypeError('basevol must be an integer')
        if basevol < 0:
            raise ValueError('basevol cannot be negative')
        self._basevol = basevol

    def pipeline(self):

        if self.func is None:
            raise RuntimeError(
                f'func must be set prior to {type(self).__name__}.run()')

        if self.basevol is None:
            raise RuntimeError(
                f'basevol must be set prior to {type(self).__name__}.run()')

        if self.basevol >= self._nvols:
            raise ValueError(
                f'basevol out of range for dataset {self.func} [0, {self._nvols})')

        # Check if output already exists (delete if overwriting)
        ts_out   = self._prefix + '-func.nii'       # fMRI time series
        base_out = self._prefix + '-func-base.nii'  # fMRI base volume
        mcf_out  = self._prefix + '-func-mc.nii'    # Motion-corrected time series
        mat_dir  = self._prefix + '-func-mc.mat'    # Tx matrix directory
        par_out  = self._prefix + '-func-mc.par'    # Tx parameters file
        log      = self._prefix + '-PrepFunc.log'   # Log
        outfiles = (ts_out, base_out, mcf_out, par_out, log)

        self._log.debug(f'output: {mat_dir}')
        if os.path.isdir(mat_dir):
            if self._overwrite:
                shutil.rmtree(mat_dir)
            else:
                raise IOError(f'{mat_dir} already exists')

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        wd = os.getcwd()
        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            env['FSLOUTPUTTYPE'] = 'NIFTI'
            with open(log, 'w', buffering=1) as lf:
                with tempfile.TemporaryDirectory(dir=self._scratch) as td:
                    os.chdir(td)
                    self._log.debug(f'Temporary directory: {td}')

                    # Re-orient the input data to FSL's preferred orientation
                    tf = os.path.join(td, 'tmp.nii')
                    self._cmd('fslreorient2std',
                        self._func,
                        tf.replace('.nii', ''),
                        env=env,
                        redirect=lf)

                    # Convert the input data to floating-point format
                    self._cmd('fslmaths',
                        tf,
                        ts_out.replace('.nii', ''),
                        '-odt', 'float',
                        env=env,
                        redirect=lf)

                    # Extract the base volume
                    self._cmd('fslroi',
                        ts_out,
                        base_out.replace('.nii',''),
                        str(self.basevol),
                        str(1),
                        env=env,
                        redirect=lf)

                    # Run motion correction using mcflirt
                    self._cmd('mcflirt',
                        '-in', ts_out,
                        '-out', mcf_out.replace('.nii',''),
                        '-mats',
                        '-plots',
                        '-reffile', base_out.replace('.nii',''),
                        '-spline_final',
                        env=env,
                        redirect=lf)
        except:
            if os.path.isdir(mat_dir):
                shutil.rmtree(mat_dir)
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise
        finally:
            os.chdir(wd)

###############################################################################

class FuncToT1(jurt.core.FslPipeline):
    """Co-register functional dataset to T1-weighted dataset"""

    _valid_methods = ('bbr', '6dof')


    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use FuncToT1 objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Required parameters')
        g1.add_argument('-method', required=True,
            choices=FuncToT1._valid_methods,
            help=FuncToT1.method.__doc__)
        g2 = p.add_argument_group('Options')
        g2.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the FuncToT1 pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

        # Set defaults
        self._method = None

    def __str__(self):
        return super().__str__() + (
            '\n  %-20s : %s' % ('Registration method', self._method))

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        if not isinstance(method, str):
            raise TypeError('method must be a string')
        if not method in FuncToT1._valid_methods:
            raise ValueError('method must be one of: {0:s}'.format(
                ', '.join(FuncToT1._valid_methods)))
        self._method = method

    def pipeline(self):

        if self.method is None:
            raise RuntimeError(
                f'method must be set prior to {type(self).__name__}.run()')

        # Check input datasets
        func     = jurt.core.Dataset(self._prefix + '-func-base.nii')
        t1_head  = jurt.core.Dataset(self._prefix + '-t1-nu.nii')
        t1_brain = jurt.core.Dataset(self._prefix + '-t1-nu-brain.nii')

        # Check if the output already exists (delete if overwriting)
        out  = self._prefix + '-func-base-t1.nii'
        omat = self._prefix + '-func-base-to-t1.mat'
        log  = self._prefix + '-FuncToT1.log'
        outfiles = (out, omat, log)

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        wd = os.getcwd()
        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            env['FSLOUTPUTTYPE'] = 'NIFTI'
            with open(log, 'w', buffering=1) as lf:
                with tempfile.TemporaryDirectory(dir=self._scratch) as td:
                    os.chdir(td)

                    init_out = os.path.join(td, 'tmp.nii')
                    init_mat = os.path.join(td, 'tmp.mat')

                    if self._method == '6dof':
                        init_mat = omat

                    # 6-DOF registration
                    # This serves as the "initial" registration for BBR and
                    # as the "final" registration for 6dof
                    # Specify "normal" search radius of 90 degrees b/c all data
                    # have been reoriented to FSL standard orientation

                    self._cmd('flirt',
                        '-ref'      , t1_brain,
                        '-in'       , func,
                        '-out'      , init_out.replace('.nii',''),
                        '-omat'     , init_mat,
                        '-cost'     , 'corratio',
                        '-dof'      , str(6),
                        '-searchrx' , str(-90), str(90),
                        '-searchry' , str(-90), str(90),
                        '-searchrz' , str(-90), str(90),
                        '-interp'   , 'trilinear',
                        env=env,
                        redirect=lf)

                    # BBR

                    if self._method == 'bbr':

                        bbr_out = os.path.join(td, 'bbr.nii')
                        wmseg   = jurt.Dataset(self._prefix +
                                  '-t1-nu-brain-wmseg.nii')
                        sch     = os.path.join(os.environ['FSLDIR'], 'etc',
                                  'flirtsch', 'bbr.sch')
                        if not os.path.isfile(sch):
                            raise IOError(
                                f'Could not find FLIRT BBR schedule: {sch}')

                        self._cmd('flirt',
                            '-ref'      , t1_head,
                            '-in'       , func,
                            '-dof'      , str(6),
                            '-cost'     , 'bbr',
                            '-searchrx' , str(-90), str(90),
                            '-searchry' , str(-90), str(90),
                            '-searchrz' , str(-90), str(90),
                            '-wmseg'    , wmseg,
                            '-init'     , init_mat,
                            '-omat'     , omat,
                            '-out'      , bbr_out.replace('.nii',''),
                            '-interp'   , 'trilinear',
                            '-schedule' , sch)

                    # Final output with sinc interpolation
                    # Apply the transformation in omat using sinc interpolation

                    self._cmd('applywarp',
                        '-i', func,
                        '-r', t1_head,
                        '-o', out.replace('.nii',''),
                        f'--premat={omat}',
                        '--interp=spline',
                        env=env,
                        redirect=lf)

        except:
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise

        finally:
            os.chdir(wd)

###############################################################################

class T1ToStd(jurt.core.FslPipeline):
    """Spatially normalize T1-weighted dataset to standard space

    Transform the nonuniformity-corrected T1-weighted dataset into standard
    stereotaxic space using a 12-DOF (full affine) registration.

    The standard space dataset is MNI152_T1_2mm_brain
    """

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use T1ToStd objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Options')
        g1.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the T1ToStd pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

    def __str__(self):
        return super().__str__()

    def pipeline(self):

        # Check the input datasets (template already verified by FslPipeline)
        t1_head  = jurt.core.Dataset(self._prefix + '-t1-nu.nii')
        t1_brain = jurt.core.Dataset(self._prefix + '-t1-nu-brain.nii')

        # Check if the output already exists (delete if overwriting)
        out  = self._prefix + '-t1-std.nii'
        omat = self._prefix + '-t1-to-std.mat'
        log  = self._prefix + '-T1ToStd.log'
        outfiles = (out, omat, log)

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            env['FSLOUTPUTTYPE'] = 'NIFTI'
            with open(log, 'w', buffering=1) as lf:
                self._cmd('flirt',
                    '-in'       ,  t1_brain,
                    '-ref'      , self._template,
                    '-out'      , out,
                    '-omat'     , omat,
                    '-cost'     , 'corratio',
                    '-dof'      , str(12),
                    '-searchrx' , str(-90), str(90),
                    '-searchry' , str(-90), str(90),
                    '-searchrz' , str(-90), str(90),
                    '-interp'   , 'trilinear',
                    env=env,
                    redirect=lf)

        except:
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise

###############################################################################

class FuncToStd(jurt.core.FslPipeline):
    """Spatially normalize functional dataset"""

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use FuncToStd objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        g1 = p.add_argument_group('Options')
        g1.add_argument('-help', action='help',
            help='Show this help message and exit')
        return p

    @classmethod
    def main(cls, ns):
        """Run the FuncToStd pipeline using an argparse namespace"""
        jurt.core._pipeline_main(cls, ns)

    def __init__(self):
        super().__init__()

    def __str__(self):
        return super().__str__()

    def pipeline(self):

        # Check the input datasets and matrix files
        func    = jurt.core.Dataset(self._prefix + '-func.nii')
        base2t1 = jurt.core.Dataset(self._prefix + '-func-base-to-t1.mat')
        t12std  = jurt.core.Dataset(self._prefix + '-t1-to-std.mat')

        mcmat = self._prefix + '-func-mc.mat'
        if not os.path.isdir(mcmat):
            raise IOError(f'Could not find directory: {mcmat}')

        n = nvols(func)
        for v in range (0, n):
            f = os.path.join(mcmat, 'MAT_{0:04d}'.format(v))
            if not os.path.isfile(f):
                raise IOError(f'Could not find file: {f}')

        # Check if the output already exists (delete if overwriting)
        out      = self._prefix + '-func-std.nii'
        log      = self._prefix + '-FuncToStd.log'
        base2std = self._prefix + '-func-base-to-std.mat'
        #omats    = self._prefix + '-func-to-std.mat'
        outfiles = (out, base2std, log)

        #self._log.debug(f'output {omats}')
        #if os.path.isdir(omats):
        #    if self._overwrite:
        #        shutil.rmtree(omats)
        #    else:
        #        raise IOError(f'{omats} already exists')

        for f in outfiles:
            self._log.debug(f'output {f}')
            if os.path.isfile(f):
                if self._overwrite:
                    os.remove(f)
                else:
                    raise IOError(f'{f} already exists')

        wd = os.getcwd()
        try:
            env = os.environ
            env['OMP_NUM_THREADS'] = str(self._threads)
            env['FSLOUTPUTTYPE'] = 'NIFTI'
            with open(log, 'w', buffering=1) as lf:
                with tempfile.TemporaryDirectory(dir=self._scratch) as td:
                    os.chdir(td)
                    self._log.debug(f'Temporary directory: {td}')

                    # Concatenate func-to-t1.mat and t1-to-std.mat
                    self._cmd('convert_xfm',
                            '-omat', base2std,
                            '-concat',
                            t12std,
                            base2t1,
                            env=env,
                            redirect=lf)

                    if not os.path.isfile(base2std):
                        raise IOError(f'Could not create transform {base2std}')

                    # Split the functional time series into individual volumes
                    # This is necessary because applyxfm4d does not properly set
                    # the TR on the output volume

                    base1 = os.path.join(td, 'vol')
                    base2 = os.path.join(td, 'std')
                    self._cmd('fslsplit',
                        func,
                        base1,
                        '-t',
                        env=env,
                        redirect=lf)

                    # Loop on volumes
                    for v in range (0, n):

                        # Concatenate func2base matrix for this volume and the
                        # base2std matrix to get func2std for this volume
                        funcvol   = '{0:s}{1:04d}.nii'.format(base1, v)
                        stdvol    = '{0:s}{1:04d}.nii'.format(base2, v)
                        func2base = os.path.join(mcmat, 'MAT_{0:04d}'.format(v))
                        func2std  = os.path.join(td, 'F2S_{0:04d}'.format(v))

                        if not os.path.isfile(funcvol):
                            raise IOError(f'Could not find volume {funcvol}')
                        if not os.path.isfile(func2base):
                            raise IOError(f'Could not find transform {func2base}')

                        self._cmd('convert_xfm',
                            '-omat', func2std,
                            '-concat',
                            base2std,
                            func2base,
                            env=env,
                            redirect=lf)

                        # Apply the func2std transformation to get the volume
                        # in standard space
                        self._cmd('flirt',
                            '-ref', self._template,
                            '-in',  funcvol,
                            '-applyxfm',
                            '-init', func2std,
                            '-out', stdvol,
                            '-interp', 'sinc',
                            '-sincwindow', 'blackman',
                            env=env,
                            redirect=lf)

                    # Merge the functional data
                    # The TR is properly set b/c it is preserved in the header
                    # of the individual images as pixdim[4]

                    stdfiles = [
                        '{0:s}{1:04d}.nii'.format(base2, vv) for vv in range(0, n)]


                    self._cmd('fslmerge',
                        '-t',
                        out,
                        *stdfiles,
                        env=env,
                        redirect=lf)

        except:
            for f in outfiles:
                if os.path.isfile(f):
                    os.remove(f)
            raise
        finally:
            os.chdir(wd)

###############################################################################

if __name__ == '__main__':
    raise RuntimeError('jurt/reg.py cannot be directly executed')

###############################################################################




"""Unit test entry point for jurt"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import sys
import os
import unittest

def main():
    """Discover and run jurt unit tests"""
    loader = unittest.TestLoader()
    suite = loader.discover(os.path.abspath(os.path.dirname(__file__)))
    runner = unittest.TextTestRunner()
    runner.run(suite)

if __name__ == '__main__':
    sys.exit(main())




# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import sys
import os
import io
import setuptools

def main():
    """Configure and build the jurt package"""

    # Set the version string
    # This is automatically updated by bumpversion (see .bumpversion.cfg)
    version = '1.0.1'

    # Dependency lists
    python_requires='>=3.6'
    install_requires=[]

    # Set package metadata
    name = 'jurt'
    description = "Jeff's Unified Registration Tool"
    jme = 'Jeffrey M. Engelmann'
    jme_email = 'jme2041@icloud.com'
    url = 'https://github.com/jme2041/jurt'

    # Get the package's long description from README.md
    here = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read().replace(
            '(LICENSE.txt)',
            '(%s/blob/master/LICENSE.txt)' % url)
    long_description_content_type = 'text/markdown'

    # Entry points
    entry_points = { 'console_scripts': 'jurt = jurt.__main__:main' }

    # Call setuptools.setup for the package
    setuptools.setup(
        name=name,
        version=version,
        description=description,
        long_description=long_description,
        long_description_content_type=long_description_content_type,
        license='BSD 3-Clause License',
        url=url,
        author=jme,
        author_email=jme_email,
        maintainer=jme,
        maintainer_email=jme_email,
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'License :: OSI Approved :: BSD License',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: Implementation :: CPython',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX :: Linux',
            'Environment :: Console',
            'Natural Language :: English',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Bio-Informatics'
        ],
        keywords='mri fmri analysis preprocessing registration normalization',
        packages=setuptools.find_packages(exclude=['test']),
        test_suite='test.get_suite',
        python_requires=python_requires,
        install_requires=install_requires,
        entry_points=entry_points)

if __name__ == '__main__':
    sys.exit(main())




"""Core jurt functionality

Support basic processing pipelines.
"""
# jurt: Jeff's Unified Registration Tool
#
# Copyright (c) 2018, Jeffrey M. Engelmann
#
# jurt is released under the revised (3-clause) BSD license.
# For details, see LICENSE.txt
#

import sys
import os
import re
import subprocess
import argparse
import logging

from jurt import __version__

###############################################################################

def _pipeline_main(cls, ns):

    # Run pipeline from class cls using argparse namespace ns

    delattr(ns, 'main')
    d = vars(ns)
    obj = cls()
    for arg in d:
        setattr(obj, arg, d[arg])
    obj.run()

###############################################################################

class Prefix(str):
    """Dataset prefix (including path)"""

    def __new__(cls, prefix):
        if not isinstance(prefix, str):
            raise TypeError('prefix must be a string')
        if len(prefix) == 0:
            raise ValueError('prefix cannot be empty')
        ap = os.path.abspath(prefix)
        wd = os.path.dirname(ap)
        if not os.path.isdir(wd):
            raise IOError(f'Could not find directory: {wd}')
        return str.__new__(cls, ap)

###############################################################################

class Dataset(str):
    """Full dataset name (including path)"""

    def __new__(cls, dataset):
        if not isinstance(dataset, str):
            raise TypeError('dataset name must be a string')
        if len(dataset) == 0:
            raise ValueError('dataset name cannot be empty')
        ap = os.path.abspath(dataset)
        if not os.path.isfile(ap):
            raise IOError(f'Could not find dataset: {ap}')
        return str.__new__(cls, ap)

###############################################################################

class Pipeline(object):
    """Processing pipeline"""

    _default_threads = 1
    _default_umask = 0o077

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use Pipeline objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False)
        g1 = p.add_argument_group('Required parameters')
        g1.add_argument('-prefix', required=True,
            help=Pipeline.prefix.__doc__)
        g2 = p.add_argument_group('Options')
        g2.add_argument('-overwrite', action='store_true',
            help=Pipeline.overwrite.__doc__)
        g2.add_argument('-scratch', metavar='DIR',
            help=Pipeline.scratch.__doc__)
        g2.add_argument('-threads', metavar='n', type=int,
            default=Pipeline._default_threads,
            help=Pipeline.threads.__doc__)
        g2.add_argument('-umask', metavar='m',
            type=lambda m: int(m,8),
            default=Pipeline._default_umask,
            help=Pipeline.umask.__doc__)
        return p

    def __init__(self):

        # Set defaults
        self._prefix    = None
        self._overwrite = False
        self._scratch   = None
        self._threads   = Pipeline._default_threads
        self._umask     = Pipeline._default_umask

        # jurt logger
        self._log = logging.getLogger(__name__)
        self._log.debug('Python version: %d.%d.%d.%s%d' % sys.version_info)
        self._log.debug(f'jurt version: {__version__}')

    def __str__(self):
        return (
            f'jurt.{type(self).__name__} object at {id(self):#x}\n' +
            '  %-20s : %s\n'  % ('Prefix', self._prefix) +
            '  %-20s : %s\n'  % ('Overwrite', self._overwrite) +
            '  %-20s : %s\n'  % ('Scratch space', self._scratch) +
            '  %-20s : %d\n'  % ('Threads requested', self._threads) +
            '  %-20s : %#05o' % ('Umask', self._umask))

    @property
    def prefix(self):
        """Output prefix (including path)"""
        return self._prefix

    @prefix.setter
    def prefix(self, prefix):
        self._prefix = Prefix(prefix)

    @property
    def overwrite(self):
        """Overwrite existing output"""
        return self._overwrite

    @overwrite.setter
    def overwrite(self, overwrite):
        if not isinstance(overwrite, bool):
            raise TypeError('overwrite must be boolean')
        self._overwrite = overwrite

    @property
    def scratch(self):
        """Scratch directory"""
        return self._scratch

    @scratch.setter
    def scratch(self, scratch):
        if scratch is None:
            self._scratch = None
            return
        if not isinstance(scratch, str):
            raise TypeError('scratch must be a string or None')
        if len(scratch) == 0:
            raise ValueError('scratch cannot be empty')
        ap = os.path.abspath(scratch)
        if not os.path.isdir(ap):
            raise IOError(f'Could not find directory: {ap}')
        self._scratch = ap

    @property
    def threads(self):
        """Number of threads requested"""
        return self._threads

    @threads.setter
    def threads(self, threads):
        if type(threads) is not int:
            raise TypeError('threads must be an integer')
        if threads <= 0:
            raise ValueError('threads must be greater than zero')
        self._threads = threads

    @property
    def umask(self):
        """Permissions mask (umask) requested"""
        return self._umask

    @umask.setter
    def umask(self, umask):
        if type(umask) is not int:
            raise TypeError('umask must be an integer')
        if umask < 0 or umask > 0o777:
            raise ValueError(
                'umask must be in the range [{0:#05o},{1:#05o}]'.format(0, 0o777))
        self._umask = umask

    def _cmd(self, *cmd, env=None, redirect=subprocess.DEVNULL):
        self._log.info(' '.join(cmd))
        if redirect != subprocess.DEVNULL:
            print(' '.join(cmd), file=redirect)
            redirect.flush()
        sp = subprocess.Popen(cmd, env=env,
            stdout=redirect, stderr=subprocess.STDOUT)
        sp.communicate()
        rc = sp.returncode
        if rc != 0:
            raise RuntimeError(f'{cmd[0]} failed with code {rc}')

    def run(self):
        """Run the processing pipeline"""

        self._log.info(f'{type(self).__name__} begins')
        old_umask = None

        try:

            # All pipelines require a prefix
            if self.prefix is None:
                raise RuntimeError(
                    f'prefix must be set prior to {type(self).__name__}.run()')

            # Set the umask
            old_umask = os.umask(self._umask)

            # Run the pipeline
            self.pipeline()

        except Exception as e:
            self._log.exception(e)
            raise

        finally:
            if old_umask is not None:
                os.umask(old_umask)
            self._log.info(f'{type(self).__name__} ends')

    def pipeline(self):
        """Processing pipeline"""
        raise NotImplementedError(
            f'{type(self).__name__}.pipeline() must be overridden')

###############################################################################

class FsPipeline(Pipeline):
    """FreeSurfer processing pipeline"""

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use FsPipeline objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        return p

    def __init__(self):
        super().__init__()

        # Check FreeSurfer version
        FS_MIN = (6, 0, 0)
        cmd = subprocess.Popen(['recon-all', '-version'], stdout=subprocess.PIPE)
        o, e = cmd.communicate()
        if cmd.returncode != 0:
            raise RuntimeError('Could not determine FreeSurfer version')
        fs_verstr = o.decode().splitlines()[0]
        pattern = '^freesurfer.+\-v(\d+)\.(\d+)\.(\d+)\-\w+$'
        match = re.search(pattern, fs_verstr)
        if match is None:
            raise RuntimeError('Could not determine FreeSurfer version')
        self._fs_ver = tuple(map(int, match.groups()))
        self._log.debug('FreeSufer version %d.%d.%d' % (self._fs_ver))
        if self._fs_ver < FS_MIN:
            raise RuntimeError(
                    'FreeSurfer %d.%d.%d or newer is required\n' % (FS_MIN) +
                    'Found version %d.%d.%d' % (self._fs_ver))

        # Set default gca
        self._gca = os.path.join(os.environ['FREESURFER_HOME'], 'average',
            'RB_all_withskull_2016-05-10.vc700.gca')
        if not os.path.isfile(self._gca):
            raise IOError(f'Could not find gca template: {self._gca}')

    def __str__(self):
        return super().__str__() + '\n' + (
            '  %-20s : ' % 'FreeSurfer version' + '%d.%d.%d\n' % self._fs_ver +
            '  %-20s : ' % 'Default gca' + os.path.basename(self._gca))

###############################################################################

class FslPipeline(Pipeline):
    """FSL processing pipeline"""

    @classmethod
    def parser(cls):
        """Return an argument parser for scripts that use FslPipeline objects"""
        p = argparse.ArgumentParser(add_help=False, allow_abbrev=False,
            parents=[cls.__base__.parser()])
        return p


    def __init__(self):
        super().__init__()

        # Check FSL version via FAST
        FSL_MIN = (5, 0, 11)
        cmd = subprocess.Popen(['fast', '--help'], stderr=subprocess.PIPE)
        o, e = cmd.communicate()
        if cmd.returncode != 1:
            raise RuntimeError('Could not determine FSL version via FAST')
        fast_verstr = e.decode().splitlines()[1]
        pattern = '^Part\sof\sFSL\s\(ID:\s(\d+)\.(\d+)\.(\d+)\)$'
        match = re.search(pattern, fast_verstr)
        if match is None:
            raise RuntimeError('Could not determine FSL version via FAST')
        self._fsl_ver = tuple(map(int, match.groups()))
        self._log.debug('FSL version %d.%d.%d' % (self._fsl_ver))
        if self._fsl_ver < FSL_MIN:
            raise RuntimeError(
                'FSL %d.%d.%d or newer is required\n' % (FSL_MIN) +
                'Found version %d.%d.%d' % (self._fsl_ver))

        # Check FLIRT version
        FLIRT_MIN = (6, 0)
        cmd = subprocess.Popen(['flirt', '-version'], stdout=subprocess.PIPE)
        o, e = cmd.communicate()
        if cmd.returncode != 0:
            raise RuntimeError('Could not determine FLIRT version')
        flirt_verstr = o.decode().splitlines()[0]
        pattern = '^FLIRT\s+version\s+(\d+)\.(\d+).*$'
        match = re.search(pattern, flirt_verstr)
        if match is None:
            raise RuntimeError('Could not determine FLIRT version')
        self._flirt_ver = tuple(map(int, match.groups()))
        self._log.debug('FEAT version %d.%d' % (self._flirt_ver))
        if self._flirt_ver < FLIRT_MIN:
            raise RuntimeError(
                'FLIRT %d.%d or newer is required\n' % (FLIRT_MIN) +
                'Found version %d.%d' % (self._flirt_ver))

        # Set default standard-space image
        self._template = os.path.join(os.environ['FSLDIR'], 'data', 'standard',
            'MNI152_T1_2mm_brain.nii.gz')
        if not os.path.isfile(self._template):
            raise IOError(f'Could not find standard space dataset: {self._template}')

    def __str__(self):
        return super().__str__() + '\n' + (
            '  %-20s : ' % 'FSL version'       + '%d.%d.%d\n' % self._fsl_ver +
            '  %-20s : ' % 'FSL FLIRT version' + '%d.%d\n' % self._flirt_ver +
            '  %-20s : ' % 'Default template' + os.path.basename(self._template))

###############################################################################

if __name__ == '__main__':
    raise RuntimeError('jurt/core.py cannot be directly executed')

###############################################################################

